@@ -11,8 +11,16 @@ from urllib3.util.retry import Retry
 from concurrent.futures import ThreadPoolExecutor, as_completed
 from sqlalchemy.types import VARCHAR
 from sqlalchemy.types import DateTime
+from sqlalchemy import MetaData,Column, Table
+from sqlalchemy.orm import sessionmaker, mapper
+from sqlalchemy.ext.declarative import declarative_base
 from sqlalchemy.schema import Index
 from threading import Thread
+from sqlalchemy.exc import ArgumentError
+from sqlalchemy.sql import case
+import sqlalchemy
+
+
 import re
 from typing import TYPE_CHECKING
 if TYPE_CHECKING:
@@ -127,6 +135,17 @@ class Dataset():
         #return an array
         return dataset_choose.id.values
 
+    def get_dataset_id(self)->str:
+        """
+            get dataset id from the dataset instantiated with slug
+        :return: dataset_id: str
+        """
+        if hasattr(self,'_dataset_id'):
+            return self._dataset_id
+
+        else:
+            raise Exception("dataset didn't instantiate with slug")
+
     # TODO: Cancellare dopo aver verificato che load_datset_to_df va ok perchè possiamo riutilizzare __get_dataset_settings_info per avere le stesse informazione di return
 
     # def __get_dataset_connection(self, datasetid):
@@ -176,14 +195,23 @@ class Dataset():
 
         return entitie_code, columnsid
 
-    def execute_dataset_sync(self, dataset_id):
+    def execute_dataset_sync(self, dataset_id=''):
+
+        if dataset_id != '':
+            pass
+        elif hasattr(self, '_dataset_slug') != '' and dataset_id == '':
+            dataset_id = self._dataset_id
+            logging.debug("---- sync dataset with id '{}' ----- ".format(str(self._dataset_id)))
+        else:
+            raise Exception("takes 2 positional arguments but dataset_id weren't given or dataset didn't"
+                            " instantiate with slug")
 
         dataset_url = self._base_url_askdata + '/smartdataset/datasets/' + dataset_id + '/sync'
         r = requests.post(url=dataset_url, headers=self._headers)
         r.raise_for_status()
         return r
 
-    def __ask_db_engine(self, dataset_id, setting):
+    def __ask_db_engine(self, dataset_id: str, setting: dict):
         # request credential
 
         s = requests.Session()
@@ -223,10 +251,11 @@ class Dataset():
         logging.debug('-------delete mysqluser for dataset {}------'.format(dataset_id))
 
 
-    def save_to_dataset(self, frame: pd.DataFrame, dataset_name: str, add_indexdf = False,
-                        indexclm = [], if_exists='Replace') -> str:
+    def create_dataset(self, frame: pd.DataFrame, dataset_name: str, add_indexdf = False,
+                       indexclm = [], unique_key=[]) -> str:
 
         # TODO: see upsert in mysql if_exists['replace','Append','upsert']
+        # TODO: insert unique_key
         '''
         Save the data frame in askdata dataset of the specific agent
 
@@ -294,7 +323,11 @@ class Dataset():
         # delete mysql user
         self.__ask_del_db_engine(dataset_id)
 
-        return dataset_id
+        # find list dataset
+        list_dataset = self.list_datasets()
+        slug = list_dataset[list_dataset['id'] == dataset_id].loc[:,'slug'].item()
+
+        return slug
 
 
 
@@ -327,7 +360,7 @@ class Dataset():
 
         #TODO: develop support for any dataset type
 
-        if dataset_info["type"] != 'MYSQL' or dataset_info["type"] != 'DATAFRAME':
+        if dataset_info["type"] != 'MYSQL' and dataset_info["type"] != 'DATAFRAME':
             raise Exception('dataset {} not support for loading in dataframe'.format(str(dataset_info["type"])))
 
         table_id = dataset_info["settings"]["table_id"]
@@ -582,7 +615,7 @@ class Dataset():
         logging.info('Dataset {} copied in the agent'.format(label))
 
         # retrive all settings of the entities (source dataset)
-        dataset_entities_doc_source = agent_source.retrive_dataset_entities(dataset_id=dataset_id_source, dataset_type=type_dataset)
+        dataset_entities_doc_source = agent_source.__retrive_dataset_entities(dataset_id=dataset_id_source, dataset_type=type_dataset)
 
         for index, entity in enumerate(dataset_entities_doc_source["data"]):
             self.copy_entity_dataset(entity_code=entity['code'], dataset_id_dest=dataset_id_dest, settigs_entity_source=entity,
@@ -637,7 +670,7 @@ class Dataset():
 
         return dataset_document
 
-    def retrive_dataset_entities(self, dataset_id: str, dataset_type: str) -> dict:
+    def __retrive_dataset_entities(self, dataset_id: str, dataset_type: str) -> dict:
 
         """
         return a dict with all settings of the datatset's entities
@@ -691,7 +724,7 @@ class Dataset():
             response.raise_for_status()
 
         # retrive setting of destination entity
-        settigs_entities_dest = self.retrive_dataset_entities(dataset_id=dataset_id_dest, dataset_type=dataset_type)
+        settigs_entities_dest = self.__retrive_dataset_entities(dataset_id=dataset_id_dest, dataset_type=dataset_type)
 
         # insert settings of source entity in destination entity
         for index, settigs_entity_dest in enumerate(settigs_entities_dest['data']):
@@ -722,6 +755,14 @@ class Dataset():
                                               settigs_entity["parameterType"], dataset_type)
 
     def copy_values_entity_dataset(self, entity_code: str, dataset_id_dest: str, entity_values_settings_source_list: list):
+        """
+        copy a list of settings of values in a new the entity_code of the destination dataset id
+
+        :param entity_code: str, entity_code of the destination dataset
+        :param dataset_id_dest: str, id dataset of the destination dataset
+        :param entity_values_settings_source_list: list, list of dict of the settings values
+        :return: None
+        """
 
         # retrive the setting's values of specific entity code
         entity_values_settings_dest = self.__get_value_entity(entity_code=entity_code)
@@ -754,6 +795,159 @@ class Dataset():
                                             settings_value=value_entity)
 
 
+    def get_columns_code(self)->list:
+        """
+        return the list of column of specific dataset instantiated with slug
+        :return: list
+        """
+
+        if hasattr(self, '_dataset_id'):
+            s = requests.Session()
+            s.keep_alive = False
+            retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
+            s.mount('https://', HTTPAdapter(max_retries=retries))
+
+            authentication_url = self._base_url_askdata + '/smartdataset/datasets/' + self._dataset_id + '/datasetParameters/codes/list'
+            response = s.get(url=authentication_url, headers=self._headers)
+            response.raise_for_status()
+
+            column_list = [d['id'] for d in response.json()]
+            return column_list
+
+        else:
+            raise Exception("dataset didn't instantiate with slug")
+
+
+
+    def set_synonym(self,column_code: str,synonyms: list, replace=False):
+        """
+        set the synonym of column_code of specific dataset instantiated with slug
+
+        :param column_code: str, code of the entity
+        :param synonyms: list of string
+        :param replace: bool, the default is False if the Replace value is False, synonyms are added to existing
+                        synonyms otherwise they are replaced
+        :return: None
+        """
+        column_code_settings = self.__retrive_entity(entity_code=column_code, dataset_id=self._dataset_id,
+                                                     dataset_type=self._dataset_type)
+        if replace:
+            column_code_settings['synonyms'] = synonyms
+        else:
+            column_code_settings['synonyms'].extend(synonyms)
+            column_code_settings['synonyms'] = list(set(column_code_settings['synonyms']))
+
+        self.__put_entity_dataset(entity_code=column_code_settings['code'], dataset_id=self._dataset_id,
+                                  settigs_entity=column_code_settings,
+                                  entity_type=column_code_settings['parameterType'],dataset_type=self._dataset_type)
+
+
+    def get_synonym(self,column_code: str) -> list:
+        """
+        get the synonym of column_code of specific dataset instantiated with slug
+
+        :param column_code: str
+        :return: list, list of the synonym
+        """
+
+        column_code_settings = self.__retrive_entity(entity_code=column_code, dataset_id=self._dataset_id,
+                                                     dataset_type=self._dataset_type)
+        return column_code_settings.get('synonyms',[])
+
+    def del_synonym(self,column_code: str, syn: list) :
+        """
+        delete the synonym of column_code of specific dataset instantiated with slug
+
+        :param column_code: str value of entity
+        :param syn: str value of synonym to delete
+        :return: list, list of the synonym
+        """
+        synonyms = self.get_synonym(column_code)
+
+        # check if synonyms is empty
+        if len(synonyms) > 0:
+            for syn_el in syn:
+                synonyms.remove(syn_el)
+
+        self.set_synonym(column_code=column_code, synonyms=synonyms, replace=True)
+
+    def set_injections(self,column_code: str,injections: list, replace=False):
+        """
+        set the colun code to injection of specific dataset instantiated with slug
+
+        :param column_code: str, code of the entity
+        :param injections: list of column_code to injection
+        :param replace: bool, the default is False if the Replace value is False, injections are added to existing
+                        injections otherwise they are replaced
+        :return: None
+        """
+        column_code_settings = self.__retrive_entity(entity_code=column_code, dataset_id=self._dataset_id,
+                                                     dataset_type=self._dataset_type)
+        if replace:
+            column_code_settings['advancedConfiguration']['injections'] = injections
+        else:
+            if column_code_settings['advancedConfiguration'].get('injections') != None:
+
+                column_code_settings['advancedConfiguration']['injections'].extend(injections)
+                column_code_settings['advancedConfiguration']['injections'] = \
+                    list(set(column_code_settings['advancedConfiguration']['injections']))
+
+            else:
+                column_code_settings['advancedConfiguration'].setdefault('injections', injections)
+
+
+        self.__put_entity_dataset(entity_code=column_code_settings['code'], dataset_id=self._dataset_id,
+                                  settigs_entity=column_code_settings,
+                                  entity_type=column_code_settings['parameterType'],dataset_type=self._dataset_type)
+
+    def get_injections(self, column_code: str) -> list:
+            """
+            get the injections of column_code of specific dataset instantiated with slug
+
+            :param column_code: str
+            :return: list, list of the injections
+            """
+
+            column_code_settings = self.__retrive_entity(entity_code=column_code, dataset_id=self._dataset_id,
+                                                         dataset_type=self._dataset_type)
+            return column_code_settings['advancedConfiguration'].get('injections',[])
+
+    def del_injections(self,column_code: str, injection_list: list) :
+        """
+        delete the injections of column_code of specific dataset instantiated with slug
+
+        :param column_code: str value of entity
+        :param injection: str value of synonym to delete
+        :return: list, list of the synonym
+        """
+        injections = self.get_injections(column_code)
+
+        # check for injection list empty
+        if len(injections)>0:
+            for injection in injection_list:
+                injections.remove(injection)
+
+        self.set_injections(column_code=column_code, injections=injections, replace=True)
+
+    def __retrive_entity(self,entity_code: str, dataset_id: str, dataset_type: str)-> dict:
+        """
+        retrive the entity settings for a specific entity_code and dataset
+
+        :param entity_code: str
+        :param dataset_id_dest: str
+        :param dataset_type: str
+        :return: dict
+        """
+        # retrive setting of destination entity
+        settigs_entities_dest = self.__retrive_dataset_entities(dataset_id=dataset_id, dataset_type=dataset_type)
+
+        # insert settings of source entity in destination entity
+        for index, settigs_entity_dest in enumerate(settigs_entities_dest['data']):
+            if settigs_entity_dest["code"] == entity_code:
+                entity_settings = settigs_entity_dest
+                break
+
+        return entity_settings
 
     def __put_entity_dataset(self, entity_code: str, dataset_id: str, settigs_entity: dict, entity_type: str, dataset_type: str):
 
@@ -964,3 +1158,60 @@ class Dataset():
         self.__put_settings_dataset(dataset_id=self._dataset_id, all_settings=dataset_document)
 
         logging.info('---- settings updated ----')
+
+    def update_dataset(self, dataframe: pd.DataFrame, dataset_slug:str ='', type_update:str='replace'):
+        """
+        update records stored in a dataset DataFrame.
+
+        :param type_update: str, {‘replace’, ‘append’}, default ‘replace’
+
+
+            replace: Drop the table before inserting new values.
+            append: Insert new values to the existing table.
+        :param unique_key: str, default ''
+            Primary key selected necessary for updating if type_update is 'update'
+        :return:
+        """
+
+        if dataset_slug != '':
+            self._get_info_dataset_by_slug(dataset_slug)
+            logging.debug("---- get info for dataset with slug '{}' ----- ".format(dataset_slug))
+
+        elif hasattr(self,'_dataset_slug') != '' and dataset_slug == '':
+            pass
+            logging.debug("---- get info for dataset with slug '{}' ----- ".format(str(self._dataset_slug)))
+        else:
+            raise Exception("takes 4 positional arguments and but dataset_slug weren't given or dataset didn't"
+                            " instantiate with slug")
+
+        #check type of dataset
+        if self._dataset_type != 'MYSQL' and self._dataset_type != 'DATAFRAME':
+            raise Exception('update is active only dataset of type DataFrame')
+
+        dataset_document = self.__get_dataset_settings_info(dataset_id=self._dataset_id)
+        engine, db_tablename = self.__ask_db_engine(dataset_id=self._dataset_id, setting=dataset_document['settings'])
+
+        with engine.connect() as connection:
+            if type_update == 'replace':
+                sql_truncate = """Truncate TABLE `{}`;""".format(db_tablename)
+                connection.execute(sql_truncate)
+
+                dataframe.to_sql(con=connection, name=db_tablename, if_exists='append', chunksize=1000, method='multi',
+                                 index=False)
+
+                self.__ask_del_db_engine(dataset_id=self._dataset_id)
+
+                logging.info('----  update dataset with id : ---'.format(self._dataset_id))
+
+            elif type_update == 'append':
+
+                dataframe.to_sql(con=connection, name=db_tablename, if_exists='append', chunksize=1000, method='multi',
+                                 index=False)
+
+                self.__ask_del_db_engine(dataset_id=self._dataset_id)
+
+                logging.info('----  update dataset with id : ---'.format(self._dataset_id))
+
+            #elif type_update == 'upsert':
+            #    pass
+            #TODO test 'replace' develop append and update and set primary key in save_to_dataset
\ No newline at end of file
